{ Unit: BcActionClasses
  ===========================================================================

  The contents of this file are subject to the Bluecave Limited License V 1.0
  (the "License"). You may not copy or use this file, in either source code
  or executable form, except in compliance with the License. You may obtain
  a copy of the License at http://www.bluecave.net/licenses/.

  Software distributed under the License is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
  for the specific language governing rights and limitations under the
  License.

  Copyright (C) 2001 Bluecave Software. All Rights Reserved.
  http://www.bluecave.net/

  Copyright (C) 2001 Jouni Airaksinen. All Rights Reserved.
  http://Mintus.Codefield.com/ -- Mintus@Codefield.com

  ===========================================================================

  Unit description:

    Bluecave Software Action classes:

  TBcAction:
    | Base Action class
    |
    | property DisableIfNoHandler: Boolean
    | property Action: TBasicAction
    |   Links this action to another action.  Category, Name and Tag
    |   properties are not linked.
    |
    | property OnTargetExecute: TTargetExecuteEvent
    |   Allows action to access the caller component through the Target: TObject
    |   parameter in the OnTargetExecute event. Note, if you have OnExecute
    |   assigned OnTargetExecute will not be called at all.


  History:

    2001-12-08: Added Data, DataObject and OwnsDataObject properties to attach
      some custom data to the TAction. These can be used in addition to the Tag
      property. If OwnsDataObject is True when the action is destroyed the object
      pointed by DataObject is freed automatically. These properties are not linked.
    2001-11-02: Added OnTargetExecute property to TBcAction. It can be
      linked to another TBcAction through it's Action property like the
      other events.
    2001-10-23: Initial version

  =========================================================================== }

{$I DFS.inc}
{$I BcDirectives.inc}

unit BcActionClasses;

interface

uses SysUtils, Classes, ActnList;

const
  DefaultOwnsDataObject = False;

type
  ECircularActionLink = class(Exception);

  TBcAction = class;

  TTargetExecuteEvent = procedure(Sender: TBcAction; Target: TObject) of object;

  TBcActionLink = class(TActionLink)
  protected
    FClient: TCustomAction;

    procedure AssignClient(AClient: TObject); override;

    function IsCaptionLinked: Boolean; override;
    function IsCheckedLinked: Boolean; override;
    function IsEnabledLinked: Boolean; override;
    function IsHelpContextLinked: Boolean; override;
    function IsHintLinked: Boolean; override;
    function IsImageIndexLinked: Boolean; override;
    function IsShortCutLinked: Boolean; override;
    function IsVisibleLinked: Boolean; override;
    function IsOnExecuteLinked: Boolean; override;

    function IsOnHintLinked: Boolean; virtual;
    function IsOnTargetExecuteLinked: Boolean; virtual;
    function IsOnUpdateLinked: Boolean; virtual;
    function IsDisableIfNoHandlerLinked: Boolean; virtual;

    procedure SetCaption(const Value: string); override;
    procedure SetChecked(Value: Boolean); override;
    procedure SetEnabled(Value: Boolean); override;
    procedure SetHelpContext(Value: THelpContext); override;
    procedure SetHint(const Value: string); override;
    procedure SetImageIndex(Value: Integer); override;
    procedure SetShortCut(Value: TShortCut); override;
    procedure SetVisible(Value: Boolean); override;
    procedure SetOnExecute(Value: TNotifyEvent); override;

    procedure SetOnTargetExecute(Value: TTargetExecuteEvent); virtual;
    procedure SetOnHint(Value: THintEvent); virtual;
    procedure SetOnUpdate(Value: TNotifyEvent); virtual;
    procedure SetDisableIfNoHandler(Value: Boolean); virtual;
  end;

  TBcActionLinkClass = class of TBcActionLink;

  TBcAction = class(TCustomAction)
  private
    FOnTargetExecute: TTargetExecuteEvent;
    FData: Integer;
    FOwnsDataObject: Boolean;
    FDataObject: TObject;
    FID: Integer;
    procedure DoActionChange(Sender: TObject);
    function IsOnTargetExecuteStored: Boolean;
    procedure SetOnTargetExecute(const Value: TTargetExecuteEvent);
    function GetID: Integer;
  protected
    FActionLink: TBcActionLink;
    procedure Loaded; override;
    procedure ActionChange(Sender: TObject; CheckDefaults: Boolean);
    function GetActionLinkClass: TBcActionLinkClass; dynamic;
    function GetDisableIfNoHandler: Boolean;
    procedure SetDisableIfNoHandler(const Value: Boolean);
    function GetOnHint: THintEvent;
    function GetOnUpdate: TNotifyEvent;
    procedure SetOnHint(const Value: THintEvent);
    procedure SetOnUpdate(const Value: TNotifyEvent);
    function GetAction: TBasicAction;
    procedure SetAction(const Value: TBasicAction);
    function IsCaptionStored: Boolean;
    function IsCheckedStored: Boolean;
    function IsEnabledStored: Boolean;
    function IsHelpContextStored: Boolean;
    function IsHintStored: Boolean;
    function IsImageIndexStored: Boolean;
    function IsShortCutStored: Boolean;
    function IsVisibleStored: Boolean;
    function IsOnExecuteStored: Boolean;
    function IsOnHintStored: Boolean;
    function IsOnUpdateStored: Boolean;
    function IsDisableIfNoHandlerStored: Boolean;
    property ActionLink: TBcActionLink read FActionLink write FActionLink;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    function HandlesTarget(Target: TObject): Boolean; override;
    procedure ExecuteTarget(Target: TObject); override;

    property DataObject: TObject read FDataObject write FDataObject;
  published
    property Caption stored IsCaptionStored;
    property Checked stored IsCheckedStored;
    property Enabled stored IsEnabledStored;
    property HelpContext stored IsHelpContextStored;
    property Hint stored IsHintStored;
    property ImageIndex stored IsImageIndexStored;
    property ShortCut stored IsShortCutStored;
    property Visible stored IsVisibleStored;
    property OnExecute stored IsOnExecuteStored;
    property OnHint: THintEvent read GetOnHint write SetOnHint stored IsOnHintStored;
    property OnUpdate: TNotifyEvent read GetOnUpdate write SetOnUpdate stored IsOnUpdateStored;

    property DisableIfNoHandler: Boolean read GetDisableIfNoHandler write SetDisableIfNoHandler stored IsDisableIfNoHandlerStored;
    property Action: TBasicAction read GetAction write SetAction;

    property ID: Integer read GetID write FID default 0;
    property Data: Integer read FData write FData default 0;
    property OwnsDataObject: Boolean read FOwnsDataObject write FOwnsDataObject default DefaultOwnsDataObject;

    property OnTargetExecute: TTargetExecuteEvent read FOnTargetExecute write SetOnTargetExecute stored IsOnTargetExecuteStored;
  end;

implementation

const
  SCircularActionLink = '%s.Action cannot be linked to %s. Circular links not allowed.';

{ TBcActionLink }

procedure TBcActionLink.AssignClient(AClient: TObject);
begin
  FClient := AClient as TCustomAction;
end;

function TBcActionLink.IsCaptionLinked: Boolean;
begin
  Result := inherited IsCaptionLinked and
    (FClient.Caption = (Action as TCustomAction).Caption);
end;

function TBcActionLink.IsCheckedLinked: Boolean;
begin
  Result := inherited IsCheckedLinked and
    (FClient.Checked = (Action as TCustomAction).Checked);
end;

function TBcActionLink.IsDisableIfNoHandlerLinked: Boolean;
begin
//  Result := Action is TCustomAction;
  Result := (Action is TBcAction) and
    (FClient.DisableIfNoHandler = (Action as TBcAction).DisableIfNoHandler);
end;

function TBcActionLink.IsEnabledLinked: Boolean;
begin
  Result := inherited IsEnabledLinked and
    (FClient.Enabled = (Action as TCustomAction).Enabled);
end;

function TBcActionLink.IsHelpContextLinked: Boolean;
begin
  Result := inherited IsHelpContextLinked and
    (FClient.HelpContext = (Action as TCustomAction).HelpContext);
end;

function TBcActionLink.IsHintLinked: Boolean;
begin
  Result := inherited IsHintLinked and
    (FClient.Hint = (Action as TCustomAction).Hint);
end;

function TBcActionLink.IsImageIndexLinked: Boolean;
begin
  Result := inherited IsImageIndexLinked and
    (FClient.ImageIndex = (Action as TCustomAction).ImageIndex);
end;

function TBcActionLink.IsOnExecuteLinked: Boolean;
begin
  Result := inherited IsOnExecuteLinked and
    (@FClient.OnExecute = @Action.OnExecute);
end;

function TBcActionLink.IsOnHintLinked: Boolean;
begin
  Result := (Action is TCustomAction) and
    (@FClient.OnHint = @(Action as TCustomAction).OnHint);
end;

function TBcActionLink.IsOnTargetExecuteLinked: Boolean;
begin
  Result := (Action is TBcAction) and
    (@(FClient as TBcAction).OnTargetExecute = @(Action as TBcAction).OnTargetExecute);
end;

function TBcActionLink.IsOnUpdateLinked: Boolean;
begin
  Result := (Action is TCustomAction) and
    (@FClient.OnUpdate = @Action.OnUpdate);
end;

function TBcActionLink.IsShortCutLinked: Boolean;
begin
  Result := inherited IsShortCutLinked and
    (FClient.ShortCut = (Action as TCustomAction).ShortCut);
end;

function TBcActionLink.IsVisibleLinked: Boolean;
begin
  Result := inherited IsVisibleLinked and
    (FClient.Visible = (Action as TCustomAction).Visible);
end;

procedure TBcActionLink.SetCaption(const Value: string);
begin
  if IsCaptionLinked then FClient.Caption := Value;
end;

procedure TBcActionLink.SetChecked(Value: Boolean);
begin
  if IsCheckedLinked then FClient.Checked := Value;
end;

procedure TBcActionLink.SetDisableIfNoHandler(Value: Boolean);
begin
  if IsDisableIfNoHandlerLinked then FClient.DisableIfNoHandler := Value;
end;

procedure TBcActionLink.SetEnabled(Value: Boolean);
begin
  if IsEnabledLinked then FClient.Enabled := Value;
end;

procedure TBcActionLink.SetHelpContext(Value: THelpContext);
begin
  if IsHelpContextLinked then FClient.HelpContext := Value;
end;

procedure TBcActionLink.SetHint(const Value: string);
begin
  if IsHintLinked then FClient.Hint := Value;
end;

procedure TBcActionLink.SetImageIndex(Value: Integer);
begin
  if IsImageIndexLinked then FClient.ImageIndex := Value;
end;

procedure TBcActionLink.SetOnExecute(Value: TNotifyEvent);
begin
  if IsOnExecuteLinked then FClient.OnExecute := Value;
end;

procedure TBcActionLink.SetOnHint(Value: THintEvent);
begin
  if IsOnHintLinked then FClient.OnHint := Value;
end;

procedure TBcActionLink.SetOnTargetExecute(Value: TTargetExecuteEvent);
begin
  if IsOnTargetExecuteLinked then (FClient as TBcAction).OnTargetExecute := Value;
end;

procedure TBcActionLink.SetOnUpdate(Value: TNotifyEvent);
begin
  if IsOnUpdateLinked then FClient.OnUpdate := Value;
end;

procedure TBcActionLink.SetShortCut(Value: TShortCut);
begin
  if IsShortCutLinked then FClient.ShortCut := Value;
end;

procedure TBcActionLink.SetVisible(Value: Boolean);
begin
  if IsVisibleLinked then FClient.Visible := Value;
end;

{ TBcAction }

function TBcAction.GetAction: TBasicAction;
begin
  if FActionLink <> nil then
    Result := FActionLink.Action
  else
    Result := nil;
end;

function TBcAction.GetActionLinkClass: TBcActionLinkClass;
begin
  Result := TBcActionLink;
end;

function TBcAction.GetDisableIfNoHandler: Boolean;
begin
  Result := inherited DisableIfNoHandler;
end;

function TBcAction.IsCaptionStored: Boolean;
begin
  Result := (ActionLink = nil) or not FActionLink.IsCaptionLinked;
end;

function TBcAction.IsCheckedStored: Boolean;
begin
  Result := (ActionLink = nil) or not FActionLink.IsCheckedLinked;
end;

function TBcAction.IsDisableIfNoHandlerStored: Boolean;
begin
  Result := (ActionLink = nil) or not FActionLink.IsDisableIfNoHandlerLinked;
end;

function TBcAction.IsEnabledStored: Boolean;
begin
  Result := (ActionLink = nil) or not FActionLink.IsEnabledLinked;
end;

function TBcAction.IsHelpContextStored: Boolean;
begin
  Result := (ActionLink = nil) or not FActionLink.IsHelpContextLinked;
end;

function TBcAction.IsHintStored: Boolean;
begin
  Result := (ActionLink = nil) or not FActionLink.IsHintLinked;
end;

function TBcAction.IsImageIndexStored: Boolean;
begin
  Result := (ActionLink = nil) or not FActionLink.IsImageIndexLinked;
end;

function TBcAction.IsOnExecuteStored: Boolean;
begin
  Result := (ActionLink = nil) or not FActionLink.IsOnExecuteLinked;
end;

function TBcAction.IsOnHintStored: Boolean;
begin
  Result := (ActionLink = nil) or not FActionLink.IsOnHintLinked;
end;

function TBcAction.IsOnUpdateStored: Boolean;
begin
  Result := (ActionLink = nil) or not FActionLink.IsOnUpdateLinked;
end;

function TBcAction.IsShortCutStored: Boolean;
begin
  Result := (ActionLink = nil) or not FActionLink.IsShortCutLinked;
end;

function TBcAction.IsVisibleStored: Boolean;
begin
  Result := (ActionLink = nil) or not FActionLink.IsVisibleLinked;
end;

procedure TBcAction.SetAction(const Value: TBasicAction);
var TargetAction: TBasicAction;
begin
  if Value = nil then
  begin
    FActionLink.Free;
    FActionLink := nil;
  end else
  begin
    { Check for possible circular link. If so then raise exception to abort. }
    TargetAction := Value;
    { we can use hard casting as "while" checks for the TBcAction type }
    while (TargetAction is TBcAction) and Assigned(TBcAction(TargetAction).Action) do
    begin
      TargetAction := TBcAction(TargetAction).Action;
      if TargetAction = Self then
        raise ECircularActionLink.CreateFmt(SCircularActionLink,
          [Self.Name, Value.Name]);
    end;

    if FActionLink = nil then
      FActionLink := GetActionLinkClass.Create(Self);
    FActionLink.Action := Value;
    FActionLink.OnChange := DoActionChange;
    ActionChange(Value, csLoading in Value.ComponentState);
    Value.FreeNotification(Self);
  end;
end;

procedure TBcAction.SetDisableIfNoHandler(const Value: Boolean);
var
  I: Integer;
begin
  if Value <> inherited DisableIfNoHandler then
  begin
    for I := 0 to FClients.Count - 1 do
      if TBasicActionLink(FClients[I]) is TBcActionLink then
        TBcActionLink(FClients[I]).SetDisableIfNoHandler(Value);
    inherited DisableIfNoHandler := Value;
    Change;
  end;
end;

procedure TBcAction.DoActionChange(Sender: TObject);
begin
  if Sender = Action then ActionChange(Sender, False);
end;

procedure TBcAction.ActionChange(Sender: TObject; CheckDefaults: Boolean);
begin
  if Action is TCustomAction then
    with TCustomAction(Sender) do
    begin
      if not CheckDefaults or (Self.Caption = '') then
        Self.Caption := Caption;
      if not CheckDefaults or (Self.Checked = False) then
        Self.Checked := Checked;
      if not CheckDefaults or (Self.Enabled = True) then
        Self.Enabled := Enabled;
      if not CheckDefaults or (Self.HelpContext = 0) then
        Self.HelpContext := HelpContext;
      if not CheckDefaults or (Self.Hint = '') then
        Self.Hint := Hint;
      if not CheckDefaults or (Self.ImageIndex = -1) then
        Self.ImageIndex := ImageIndex;
      if not CheckDefaults or (Self.ShortCut = scNone) then
        Self.ShortCut := ShortCut;
      if not CheckDefaults or (Self.Visible = True) then
        Self.Visible := Visible;
      if not CheckDefaults or (Self.DisableIfNoHandler = True) then
        Self.DisableIfNoHandler := DisableIfNoHandler;
      if not CheckDefaults or not Assigned(Self.OnExecute) then
        Self.OnExecute := OnExecute;
      if not CheckDefaults or not Assigned(Self.OnHint) then
        Self.OnHint := OnHint;
      if not CheckDefaults or not Assigned(Self.OnUpdate) then
        Self.OnUpdate := OnUpdate;
    end;

  { TBcAction specific }
  if Sender is TBcAction then
    with TBcAction(Sender) do
    begin
      if not CheckDefaults or not Assigned(Self.OnTargetExecute) then
        Self.OnTargetExecute := OnTargetExecute;
    end;
end;

procedure TBcAction.Loaded;
begin
  inherited Loaded;
  if Action <> nil then ActionChange(Action, True);
end;

function TBcAction.GetOnHint: THintEvent;
begin
  Result := inherited OnHint;
end;

function TBcAction.GetOnUpdate: TNotifyEvent;
begin
  Result := inherited OnUpdate;
end;

procedure TBcAction.SetOnHint(const Value: THintEvent);
var
  I: Integer;
begin
  if @Value <> @(inherited OnHint) then
  begin
    for I := 0 to FClients.Count - 1 do
      if TBasicActionLink(FClients[I]) is TBcActionLink then
        TBcActionLink(FClients[I]).SetOnHint(Value);
    inherited OnHint := Value;
    Change;
  end;
end;

procedure TBcAction.SetOnUpdate(const Value: TNotifyEvent);
var
  I: Integer;
begin
  if @Value <> @(inherited OnUpdate) then
  begin
    for I := 0 to FClients.Count - 1 do
      if TBasicActionLink(FClients[I]) is TBcActionLink then
        TBcActionLink(FClients[I]).SetOnUpdate(Value);
    inherited OnUpdate := Value;
    Change;
  end;
end;

procedure TBcAction.ExecuteTarget(Target: TObject);
begin
  { If there's OnExecute assigned we never reach this code, so
    there is no point in trying to execute that event. }
{  if Assigned(OnExecute) then
    OnExecute(Self);}
  if Assigned(FOnTargetExecute) then
    FOnTargetExecute(Self, Target);
end;

function TBcAction.HandlesTarget(Target: TObject): Boolean;
begin
  Result := True;
end;

function TBcAction.IsOnTargetExecuteStored: Boolean;
begin
  Result := (ActionLink = nil) or not FActionLink.IsOnTargetExecuteLinked;
end;

procedure TBcAction.SetOnTargetExecute(const Value: TTargetExecuteEvent);
var
  I: Integer;
begin
  if @Value <> @OnTargetExecute then
  begin
    for I := 0 to FClients.Count - 1 do
      if TBasicActionLink(FClients[I]) is TBcActionLink then
        TBcActionLink(FClients[I]).SetOnTargetExecute(Value);
    FOnTargetExecute := Value;
    Change;
  end;
end;

constructor TBcAction.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FID := 0;
  FData := 0;
  FDataObject := nil;
  FOwnsDataObject := DefaultOwnsDataObject;
end;

destructor TBcAction.Destroy;
begin
  if FOwnsDataObject then
    FreeAndNil(FDataObject);
  inherited Destroy;
end;

function TBcAction.GetID: Integer;
begin
  if FID = 0 then
    Result := Index
  else
    Result := FID;
end;

end.
