{ Unit: BcFontButton
  ===========================================================================

  The contents of this file are subject to the Bluecave Limited License V 1.0
  (the "License"). You may not copy or use this file, in either source code
  or executable form, except in compliance with the License. You may obtain
  a copy of the License at http://www.bluecave.net/licenses/.

  Software distributed under the License is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
  for the specific language governing rights and limitations under the
  License.

  Copyright (C) 2001-2002 Bluecave Software. All Rights Reserved.
  http://www.bluecave.net/

  Copyright (C) 2001-2002 Jouni Airaksinen. All Rights Reserved.
  http://Mintus.Codefield.com/ -- Mintus@Codefield.com

  ===========================================================================

  Unit description:

    Click opens font selection dialog and Font face is shown in the button
    caption.

    Note that the OnChange event is called also if the Apply button is pressed
    on the font dialog.

    You can link a TBcColorButton to the TBcFontButton using property
    ColorButtonLink to provide color selection for Font through
    TBcColorButton. If the font color is not changed in the Font dialog,
    then the color selected earlier is preserved (tip: disable font effects
    from FontDialogOptions to hide color selection dropdown box).

  History:

    2001-11-05: Initial version

  =========================================================================== }

{$I DFS.inc}
{$I BcDirectives.inc}

unit BcFontButton;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, BcColorButton;

const
  DefaultFontDialog = True;
  DefaultFontDialogOptions = [fdEffects];
  DefaultFontPreview = False;
  DefaultDevice = fdScreen;
  DefaultCaptionFormat = '%s, %dpt';

type
  TBcFontButton = class(TButton)
  private
    FFontDialog: Boolean;
    FFontPreview: Boolean;
    FFontDialogOptions: TFontDialogOptions;
    FSelectedFont: TFont;
    FMinFontSize: Integer;
    FMaxFontSize: Integer;
    FOnApply: TFDApplyEvent;
    FDevice: TFontDialogDevice;
    FOnClose: TNotifyEvent;
    FOnShow: TNotifyEvent;
    FCaptionFormat: string;
    FColorButtonLink: TBcColorButton;
    FOnChange: TNotifyEvent;
    procedure SetFontPreview(const Value: Boolean);
    procedure SetSelectedFont(const Value: TFont);
    procedure SetCaptionFormat(const Value: string);
    procedure SetColorButtonLink(const Value: TBcColorButton);
    { Private declarations }
  protected
    { Protected declarations }
    FColorButtonClick: TNotifyEvent;
    procedure Loaded; override;
    procedure OnApplyEvent(Sender: TObject; Wnd: HWND); virtual;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure OnColorButtonClickEvent(Sender: TObject);
    procedure OnSelectedFontChangeEvent(Sender: TObject);
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure OpenFontDialog; virtual;
    procedure Click; override;
    procedure UpdateCaption; virtual;
    procedure UpdateFontPreview; virtual;
    procedure DoApply(Wnd: HWND); virtual;
  published
    { Published declarations }
    property FontDialog: Boolean read FFontDialog write FFontDialog default DefaultFontDialog;
    property FontDialogOptions: TFontDialogOptions read FFontDialogOptions write FFontDialogOptions default DefaultFontDialogOptions;
    property FontPreview: Boolean read FFontPreview write SetFontPreview default DefaultFontPreview;
    property SelectedFont: TFont read FSelectedFont write SetSelectedFont;
    property MinFontSize: Integer read FMinFontSize write FMinFontSize default 0;
    property MaxFontSize: Integer read FMaxFontSize write FMaxFontSize default 0;
    property Device: TFontDialogDevice read FDevice write FDevice default DefaultDevice;
    property CaptionFormat: string read FCaptionFormat write SetCaptionFormat;
    property ColorButtonLink: TBcColorButton read FColorButtonLink write SetColorButtonLink;
    property OnApply: TFDApplyEvent read FOnApply write FOnApply;
    property OnClose: TNotifyEvent read FOnClose write FOnClose;
    property OnShow: TNotifyEvent read FOnShow write FOnShow;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

implementation

{ TBcFontButton }

procedure TBcFontButton.Click;
begin
  OpenFontDialog;
  inherited Click;
end;

constructor TBcFontButton.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FSelectedFont := TFont.Create;
  FSelectedFont.Assign(Font);
  FSelectedFont.OnChange := OnSelectedFontChangeEvent;

  FCaptionFormat := DefaultCaptionFormat;
  UpdateCaption;

  FDevice := DefaultDevice;
  FFontPreview := DefaultFontPreview;
  FFontDialog := DefaultFontDialog;
  FFontDialogOptions := DefaultFontDialogOptions;
end;

destructor TBcFontButton.Destroy;
begin
  FreeAndNil(FSelectedFont);
  inherited Destroy;
end;

procedure TBcFontButton.DoApply(Wnd: HWND);
begin
  if Assigned(FOnApply) then
    FOnApply(Self, Wnd);
end;

procedure TBcFontButton.Loaded;
begin
  inherited Loaded;
  UpdateCaption;
  UpdateFontPreview;
end;

procedure TBcFontButton.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  if Assigned(FColorButtonLink) and (AComponent = FColorButtonLink) and
    (Operation = opRemove) then
  begin
    FColorButtonLink := nil;
    FColorButtonClick := nil;
  end;
  inherited;
end;

procedure TBcFontButton.OnApplyEvent(Sender: TObject; Wnd: HWND);
begin
  SelectedFont := (Sender as TFontDialog).Font;
  DoApply(Wnd);
end;

procedure TBcFontButton.OnColorButtonClickEvent(Sender: TObject);
begin
  if Assigned(FColorButtonLink) then
    SelectedFont.Color := FColorButtonLink.Color;
  if Assigned(FColorButtonClick) then
    FColorButtonClick(Sender);
end;

procedure TBcFontButton.OnSelectedFontChangeEvent(Sender: TObject);
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

procedure TBcFontButton.OpenFontDialog;
var FD: TFontDialog;
begin
  try
    FD := TFontDialog.Create(Self);
    FD.Device := FDevice;
    FD.Options := FFontDialogOptions;
    FD.Font.Assign(FSelectedFont);
    FD.MinFontSize := FMinFontSize;
    FD.MaxFontSize := FMaxFontSize;
    FD.OnApply := OnApplyEvent;
    FD.OnClose := FOnClose;
    FD.OnShow := FOnShow;
    if FD.Execute then
      SelectedFont := FD.Font;
  finally
    FreeAndNil(FD);
  end;
end;

procedure TBcFontButton.SetCaptionFormat(const Value: string);
begin
  FCaptionFormat := Value;
  UpdateCaption;
end;

procedure TBcFontButton.SetColorButtonLink(const Value: TBcColorButton);
begin
  FColorButtonLink := Value;
  FColorButtonClick := FColorButtonLink.OnClick;
  FColorButtonLink.OnClick := OnColorButtonClickEvent;
  SelectedFont.Color := FColorButtonLink.Color;
end;

procedure TBcFontButton.SetFontPreview(const Value: Boolean);
begin
  if FFontPreview <> Value then
  begin
    FFontPreview := Value;
    UpdateFontPreview;
  end;
end;

procedure TBcFontButton.SetSelectedFont(const Value: TFont);
begin
  FSelectedFont.Assign(Value);
  UpdateCaption;
  UpdateFontPreview;
end;

procedure TBcFontButton.UpdateCaption;
begin
  if (ComponentState * [csLoading, csDestroying, csLoading]) <> [] then Exit;
  Caption := Format(CaptionFormat, [FSelectedFont.Name, FSelectedFont.Size]);
end;

procedure TBcFontButton.UpdateFontPreview;
begin
  if (ComponentState * [csLoading, csDestroying, csLoading]) <> [] then Exit;
  if FFontPreview then
    Font.Assign(FSelectedFont)
  else
    ParentFont := True;
end;

end.
