{ Unit: BcIndyHTTPFile
  ===========================================================================

  The contents of this file are subject to the Bluecave Public License V 1.1
  (the "License"). You may not copy or use this file, in either source code
  or executable form, except in compliance with the License. You may obtain
  a copy of the License at http://www.bluecave.net/licenses/.

  Software distributed under the License is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
  for the specific language governing rights and limitations under the
  License.

  Copyright (C) 2001-2002 Bluecave Software. All Rights Reserved.
  http://www.bluecave.net/

  Copyright (C) 2001-2002 Jouni Airaksinen. All Rights Reserved.
  http://Mintus.Codefield.com/ -- Mintus@Codefield.com

  ===========================================================================

  Unit description:

    Send a file as http POST. The FileName property is just for the data
    form sent to the Server/CGI. You need to manually load content of a file
    to Lines TStringList.

    By default this component handles plain text files. Good for example
    implementing feedback forms, error report sending, etc, to your
    application. If you need other types (binary) then change the ContentType
    to correct type and provide the data through OnGetPostContent event by passing
    valid stream in var Content: TStream. You can free up your stream on
    OnSendComplete event. Post will fail if you don't provide valid stream.

    property FilePost: Boolean controls if the file is sent as http FILE POST
    or just as a field (in html <textarea>) in URLEncoded format. This might
    come in handy if your server does not accept file posts (for example PHP
    configured in Safe-Mode).

    If the Server/CGI returns something it is show to user in a standard
    messagebox (no html support). To show nothing, return nothing.

    With DownloadFile() methods you can download url to a file. Either directly
    to a file or to stream for further processing in your application. Download
    supports binary files (e.g. images).

  Requirements:

    Requires Nevrona's Indy Components, http://www.nevrona.com/indy/

  History:

    2001-12-04: Added DownloadFile() methods (overloaded).
    2001-11-30: Added possibility to send the file as a field value instead of
      http file post. Also methods to download files (to stream or file).
    2001-11-27: Initial version

  =========================================================================== }

{$I DFS.inc}
{$I BcDirectives.inc}

unit BcIndyHTTPFile;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  IdAntiFreezeBase, IdAntiFreeze, IdBaseComponent, IdComponent, IdTCPConnection,
  IdTCPClient, IdHTTP, BcSendProgress, BcDownloadProgress;

resourcestring
  SSendError = 'Error while sending!';
  SDownloadError = 'Error while downloading!';
  SDialogTitle = 'Send';
  SSendDialogTitle = 'Sending';
  SDownloadDialogTitle = 'Downloading';
  SDownloadComplete = 'Download complete!';
  SDownloadDialogTransfer = 'Transfer:';
  DefaultFileName = 'BcIndyPostFile.txt';
  DefaultFieldName = 'BcIndyPostFile'; // this is the input field name
  DefaultContentType = 'text/plain';

const
  DefaultMaxFileSize = 500 * 1024; // 500kB
  DefaultFilePost = True;

type
  TBcSendCompleteEvent = procedure(Sender: TObject; Content: TStream) of object;
  TBcGetPostContentEvent = procedure(Sender: TObject; var Content: TStream) of object;
  TOpenProgressFormEvent = procedure(Sender: TObject; var CustomForm: Boolean) of object;

  TBcIndyHTTPFile = class(TComponent)
  private
    FHTTP: TIdHTTP;
    FBoundary: string;
    FSendProgressForm: TBcSendProgressForm;
    FDownloadProgressForm: TBcDownloadProgressForm;
    FLines: TStrings;
    FMaxFileSize: Cardinal;
    FDialogParent: THandle;
    FURL: string;
    FDialogTitle: string;
    FSendError: string;
    FFileName: string;
    FFieldName: string;
    FOnCloseSendProgessForm: TNotifyEvent;
    FOnOpenSendProgessForm: TOpenProgressFormEvent;
    FSendDialogTitle: string;
    FCursor: TCursor;
    FParentForm: TCustomForm;
    FFilePost: Boolean;
    FDownloadDialogTitle: string;
    FDownloadError: string;
    FOnCloseDownloadProgessForm: TNotifyEvent;
    FOnOpenDownloadProgessForm: TOpenProgressFormEvent;
    FContentType: string;
    FOnGetPostContent: TBcGetPostContentEvent;
    FOnSendComplete: TBcSendCompleteEvent;
    FDownloadComplete: string;
    FDownloadDialogTransfer: string;
    function GetMIME: string;
    procedure SetLines(const Value: TStrings);
    function GetBoundary: string;
    function GetURLEncoded: string;
    procedure SetHTTP(const Value: TIdHTTP);
    { Private declarations }
  protected
    { Protected declarations }
    FOldOnStatus: TIdStatusEvent;
    FOldOnWork: TWorkEvent;
    FOldOnWorkBegin: TWorkBeginEvent;
    FOldOnWorkEnd: TWorkEndEvent;
    Posting: Boolean;
    Downloading: Boolean;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;

    property SendProgressForm: TBcSendProgressForm read FSendProgressForm write FSendProgressForm;
    property DownloadProgressForm: TBcDownloadProgressForm read FDownloadProgressForm write FDownloadProgressForm;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure HTTPStatus(axSender: TObject; const axStatus: TIdStatus;
      const asStatusText: String);
    procedure HTTPWorkBegin(Sender: TObject; AWorkMode: TWorkMode;
      const AWorkCountMax: Integer);
    procedure HTTPWorkEnd(Sender: TObject; AWorkMode: TWorkMode);
    procedure HTTPWork(Sender: TObject; AWorkMode: TWorkMode;
      const AWorkCount: Integer);
    procedure CreateNewBoundary;

    function PostFile: Boolean;
    function DownloadFile(Target: TStream): Boolean; overload;
    function DownloadFile(FileName: string): Boolean; overload;

    property Boundary: string read GetBoundary;
    property DialogParent: THandle read FDialogParent write FDialogParent;
    property ParentForm: TCustomForm read FParentForm write FParentForm;
    property MIME: string read GetMIME;
    property URLEncoded: string read GetURLEncoded;
  published
    { Published declarations }
    property HTTP: TIdHTTP read FHTTP write SetHTTP;
    property URL: string read FURL write FURL;
    property DialogTitle: string read FDialogTitle write FDialogTitle;
    property SendDialogTitle: string read FSendDialogTitle write FSendDialogTitle;
    property DownloadDialogTitle: string read FDownloadDialogTitle write FDownloadDialogTitle;
    property DownloadDialogTransfer: string read FDownloadDialogTransfer write FDownloadDialogTransfer;
    property SendError: string read FSendError write FSendError;
    property DownloadError: string read FDownloadError write FDownloadError;
    property DownloadComplete: string read FDownloadComplete write FDownloadComplete;
    property Lines: TStrings read FLines write SetLines;
    property FileName: string read FFileName write FFileName;
    property FieldName: string read FFieldName write FFieldName;
    property Cursor: TCursor read FCursor write FCursor;
    property MaxFileSize: Cardinal read FMaxFileSize write FMaxFileSize default DefaultMaxFileSize;
    property FilePost: Boolean read FFilePost write FFilePost default DefaultFilePost;
    property ContentType: string read FContentType write FContentType;

    property OnSendComplete: TBcSendCompleteEvent read FOnSendComplete write FOnSendComplete;
    property OnGetPostContent: TBcGetPostContentEvent read FOnGetPostContent write FOnGetPostContent;
    property OnOpenSendProgessForm: TOpenProgressFormEvent read FOnOpenSendProgessForm write FOnOpenSendProgessForm;
    property OnCloseSendProgessForm: TNotifyEvent read FOnCloseSendProgessForm write FOnCloseSendProgessForm;
    property OnOpenDownloadProgessForm: TOpenProgressFormEvent read FOnOpenDownloadProgessForm write FOnOpenDownloadProgessForm;
    property OnCloseDownloadProgessForm: TNotifyEvent read FOnCloseDownloadProgessForm write FOnCloseDownloadProgessForm;
  end;

implementation

{ Copied from IdGlobal.pas and modified a bit }
function URLEncode(const psSrc: string): string;
const
  UnsafeChars = ' *#%<>&=';
var
  i: Integer;
begin
  Result := '';
  for i := 1 to Length(psSrc) do
  begin
    if (AnsiPos(psSrc[i], UnsafeChars) > 0) or (psSrc[i] < #$20) or (psSrc[i] >= #$80) then
    begin
      Result := Result + '%' + IntToHex(Ord(psSrc[i]), 2);
    end else
      Result := Result + psSrc[i];
  end;
end;

procedure TBcIndyHTTPFile.HTTPStatus(axSender: TObject;
  const axStatus: TIdStatus; const asStatusText: String);
begin
  if Assigned(FOldOnStatus) then
    FOldOnStatus(axSender, axStatus, asStatusText);

  if Assigned(SendProgressForm) then
    SendProgressForm.StatusText := asStatusText
  else if Assigned(DownloadProgressForm) then
    DownloadProgressForm.StatusText := asStatusText;
  Application.ProcessMessages;
end;

procedure TBcIndyHTTPFile.HTTPWorkBegin(Sender: TObject; AWorkMode: TWorkMode;
  const AWorkCountMax: Integer);
begin
  if Assigned(FOldOnWorkBegin) then
    FOldOnWorkBegin(Sender, AWorkMode, AWorkCountMax);

  Screen.Cursor := FCursor;
  if Assigned(SendProgressForm) then
    SendProgressForm.TransferMax := AWorkCountMax
  else if Assigned(DownloadProgressForm) then
    DownloadProgressForm.TransferMax := AWorkCountMax;
  Application.ProcessMessages;
end;

procedure TBcIndyHTTPFile.HTTPWorkEnd(Sender: TObject; AWorkMode: TWorkMode);
begin
  if Assigned(FOldOnWorkEnd) then
    FOldOnWorkEnd(Sender, AWorkMode);

  if Assigned(SendProgressForm) then
    SendProgressForm.ModalResult := mrOK
  else if Assigned(DownloadProgressForm) then
    DownloadProgressForm.ModalResult := mrOK;
  Application.ProcessMessages;
end;

procedure TBcIndyHTTPFile.HTTPWork(Sender: TObject; AWorkMode: TWorkMode;
  const AWorkCount: Integer);
begin
  { update cursor too }
  if Screen.Cursor <> FCursor then
    Screen.Cursor := FCursor;
  if Assigned(SendProgressForm) then
    SendProgressForm.TransferPos := AWorkCount
  else if Assigned(DownloadProgressForm) then
    DownloadProgressForm.TransferPos := AWorkCount;
  Application.ProcessMessages;
end;

function TBcIndyHTTPFile.GetMIME: string;
const
  CRLF = #13#10;
begin
  Result :=
    Format(
      '--%0:s' + CRLF +
      'Content-Disposition: form-data; name="MAX_FILE_SIZE"' + CRLF + CRLF +
      '%2:d' + CRLF +
      '--%0:s' + CRLF +
      'Content-Disposition: form-data; name="%3:s"; filename="%4:s"' + CRLF +
      'Content-Type: %5:s' + CRLF + CRLF +
      '%1:s' + CRLF +
      '--%0:s--' + CRLF,
      [Boundary, FLines.Text, MaxFileSize, FFieldName, FFileName, FContentType]);
end;

procedure TBcIndyHTTPFile.SetLines(const Value: TStrings);
begin
  FLines.Assign(Value);
end;

constructor TBcIndyHTTPFile.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FLines := TStringList.Create;
  FDialogTitle := SDialogTitle;
  FSendDialogTitle := SSendDialogTitle;
  FDownloadDialogTitle := SDownloadDialogTitle;
  FDownloadDialogTransfer := SDownloadDialogTransfer;
  FSendError := SSendError;
  FDownloadError := SDownloadError;
  FDownloadComplete := SDownloadComplete;
  FMaxFileSize := DefaultMaxFileSize;
  FFileName := DefaultFileName;
  FFieldName := DefaultFieldName;
  FContentType := DefaultContentType;
  FParentForm := nil;
  Downloading := False;
  Posting := False;
end;

destructor TBcIndyHTTPFile.Destroy;
begin
  if Assigned(HTTP) and not (csDesigning in ComponentState) then
    HTTP.DisconnectSocket;

  FreeAndNil(FLines);
  inherited Destroy;
end;

procedure TBcIndyHTTPFile.CreateNewBoundary;
var
  T: TTimeStamp;
begin
  { create mime boundary separator }
  T := DateTimeToTimeStamp(Now);
  FBoundary := StringOfChar('-', 26) + IntToHex(T.Time, 7) + IntToHex(T.Date, 7);
  FBoundary := LowerCase(FBoundary);
end;

function TBcIndyHTTPFile.GetBoundary: string;
begin
  Result := FBoundary;
end;

function TBcIndyHTTPFile.PostFile: Boolean;
var
  PostData, ContentSource: TStream;
  Response, ConvertStream: TStringStream;
  S: string;
  FreePostData: Boolean;
  CustomForm: Boolean;
begin
  if not Assigned(FHTTP) or Downloading then
  begin
    Result := False;
    Exit;
  end;
  try
    Posting := True;
    Screen.Cursor := FCursor; // allow user to choose the cursor e.g. for non blocking transfer

    Result := True; // assume everything goes well
    try
      FOldOnStatus := FHTTP.OnStatus;
      FOldOnWork := FHTTP.OnWork;
      FOldOnWorkBegin := FHTTP.OnWorkBegin;
      FOldOnWorkEnd := FHTTP.OnWorkEnd;

      FHTTP.OnStatus := HTTPStatus;
      FHTTP.OnWork := HTTPWork;
      FHTTP.OnWorkBegin := HTTPWorkBegin;
      FHTTP.OnWorkEnd := HTTPWorkEnd;

      Response := TStringStream.Create('');
      FreePostData := False;
      PostData := nil;
      try
        if FFilePost then
        begin
          CreateNewBoundary;
          if (FContentType = DefaultContentType) or (FContentType = '') then
          begin
            PostData := TStringStream.Create(MIME);
            FreePostData := True;
          end else
          begin
            ContentSource := nil;
            Result := Assigned(FOnGetPostContent);
            if Result then
            begin
              FOnGetPostContent(Self, ContentSource);
              PostData := ContentSource;
            end;
            Result := Assigned(ContentSource);
            FreePostData := False;
          end;
        end else
        begin
          if (FContentType = DefaultContentType) or (FContentType = '') then
          begin
            PostData := TStringStream.Create(URLEncoded);
            FreePostData := True;
          end else
          begin
            ContentSource := nil;
            Result := Assigned(FOnGetPostContent);
            if Result then
              FOnGetPostContent(Self, ContentSource);
            Result := Assigned(ContentSource);
            if Result then
            begin
              try
                { Copy the stream data to stringstream and create new stringstream
                  with the data URLEncoded. You shouldn't use huge streams as it
                  requires memory about 3-4 times the source stream size. Use the
                  http post file with big files. }
                ConvertStream := TStringStream.Create('');
                ConvertStream.CopyFrom(ContentSource, 0);
                PostData := TStringStream.Create(URLEncode(ConvertStream.DataString));
              finally
                FreeAndNil(ConvertStream);
                FreePostData := Assigned(PostData);
              end;
            end;
          end;
        end;
        if Result then
        begin
          try
            CustomForm := False;
            if Assigned(FOnOpenSendProgessForm) then
              FOnOpenSendProgessForm(Self, CustomForm);
            if not CustomForm then
            begin
              FSendProgressForm := TBcSendProgressForm.Create(FParentForm);
              FSendProgressForm.Caption := FSendDialogTitle;
              FSendProgressForm.StatusText := FSendDialogTitle + '...';
              FSendProgressForm.Show;
            end;
            try
              S := FSendError; // default value just in case
              if FFilePost then
                HTTP.Request.ContentType := 'multipart/form-data; boundary=' + Boundary
              else
                HTTP.Request.ContentType := 'application/x-www-form-urlencoded';
              HTTP.Post(FURL, PostData, Response);
            except
              on E: Exception do
              begin
                Screen.Cursor := crDefault;
                Result := False;
{$IFOPT D+}
                Windows.MessageBox(FSendProgressForm.Handle, PChar(E.Message), PChar(S), MB_OK or MB_ICONERROR);
{$ELSE}
                if (FSendError <> '') then
                  Windows.MessageBox(FSendProgressForm.Handle, PChar(FSendError), PChar(FDialogTitle), MB_OK or MB_ICONERROR);
{$ENDIF}
              end;
            end;
          finally
            if Assigned(FOnCloseSendProgessForm) then
              FOnCloseSendProgessForm(Self);
            FreeAndNil(FSendProgressForm);
          end;
        end else
          Windows.MessageBox(DialogParent, PChar(FSendError), PChar(FDialogTitle), MB_OK or MB_ICONERROR);
      finally
        if FreePostData then
          FreeAndNil(PostData);
        if Assigned(ContentSource) then
          FOnSendComplete(Self, ContentSource);
      end;
      S := Response.DataString;
    finally
      FreeAndNil(Response);

      if Assigned(FHTTP) then
      begin
        FHTTP.OnStatus := FOldOnStatus;
        FHTTP.OnWork := FOldOnWork;
        FHTTP.OnWorkBegin := FOldOnWorkBegin;
        FHTTP.OnWorkEnd := FOldOnWorkEnd;
      end;

      Posting := False;

      { transfer ok, show text returned by the script as messagebox text }
      if Result and (S <> '') then
      begin
        Screen.Cursor := crDefault;
        Windows.MessageBox(DialogParent, PChar(S), PChar(SDialogTitle), MB_OK or MB_ICONINFORMATION);
      end;
    end;
  finally
    Screen.Cursor := crDefault;
  end;
end;

procedure TBcIndyHTTPFile.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Assigned(FHTTP) and (AComponent = FHTTP) and (Operation = opRemove) then
  begin
    if not (csDesigning in ComponentState) then
      HTTP.DisconnectSocket;

    FHTTP := nil;
    FOldOnStatus := nil;
    FOldOnWork := nil;
    FOldOnWorkBegin := nil;
    FOldOnWorkEnd := nil;
  end;
end;

function TBcIndyHTTPFile.GetURLEncoded: string;
const
  CRLF = #13#10;
begin
  Result := Format('MAX_FILE_SIZE=%1:d&' + '%2:s_name=%3:s&' + '%2:s=%0:s',
     [URLEncode(FLines.Text), MaxFileSize, FFieldName, FFileName]);
end;

function TBcIndyHTTPFile.DownloadFile(Target: TStream): Boolean;
var CustomForm: Boolean;
begin
  if not Assigned(FHTTP) or not Assigned(Target) or Posting then
  begin
    Result := False;
    Exit;
  end;
  try
    Screen.Cursor := FCursor; // allow user to choose the cursor e.g. for non blocking transfer

    Result := True; // assume everything goes well
    try
      FOldOnStatus := FHTTP.OnStatus;
      FOldOnWork := FHTTP.OnWork;
      FOldOnWorkBegin := FHTTP.OnWorkBegin;
      FOldOnWorkEnd := FHTTP.OnWorkEnd;

      FHTTP.OnStatus := HTTPStatus;
      FHTTP.OnWork := HTTPWork;
      FHTTP.OnWorkBegin := HTTPWorkBegin;
      FHTTP.OnWorkEnd := HTTPWorkEnd;

      try
        CustomForm := False;
        if Assigned(FOnOpenDownloadProgessForm) then
          FOnOpenDownloadProgessForm(Self, CustomForm);
        if not CustomForm then
        begin
          FDownloadProgressForm := TBcDownloadProgressForm.Create(FParentForm);
          FDownloadProgressForm.Caption := FDownloadDialogTitle;
          FDownloadProgressForm.StatusText := FDownloadDialogTitle + '...';
          FDownloadProgressForm.LblTransfer.Caption := FDownloadDialogTransfer;
          FDownloadProgressForm.Show;
        end;
        try
          HTTP.Get(FURL, Target); // download the file
        except
          on E: Exception do
          begin
            Screen.Cursor := crDefault;
            Result := False;
{$IFOPT D+}
            Windows.MessageBox(FSendProgressForm.Handle, PChar(E.Message), PChar(S), MB_OK or MB_ICONERROR);
{$ELSE}
            if (FSendError <> '') then
              Windows.MessageBox(FSendProgressForm.Handle, PChar(FDownloadError), PChar(FDialogTitle), MB_OK or MB_ICONERROR);
{$ENDIF}
          end;
        end;
      finally
        if Assigned(FOnCloseDownloadProgessForm) then
          FOnCloseDownloadProgessForm(Self);
        FreeAndNil(FDownloadProgressForm);
      end;
    finally
      if Assigned(FHTTP) then
      begin
        FHTTP.OnStatus := FOldOnStatus;
        FHTTP.OnWork := FOldOnWork;
        FHTTP.OnWorkBegin := FOldOnWorkBegin;
        FHTTP.OnWorkEnd := FOldOnWorkEnd;
      end;

      Screen.Cursor := crDefault;
      { transfer ok, show text returned by the script as messagebox text }
      if Result and (FDownloadComplete <> '') then
        Windows.MessageBox(DialogParent, PChar(FDownloadComplete), PChar(SDialogTitle), MB_OK or MB_ICONINFORMATION);
    end;
  finally
    Screen.Cursor := crDefault;
  end;
end;

function TBcIndyHTTPFile.DownloadFile(FileName: string): Boolean;
var FS: TFileStream;
begin
//  Result := False;
  try
    FS := TFileStream.Create(FileName, fmOpenWrite or fmShareExclusive);
    Result := DownloadFile(FS);
  finally
    FreeAndNil(FS);
  end;
end;

procedure TBcIndyHTTPFile.SetHTTP(const Value: TIdHTTP);
begin
  FHTTP := Value;
end;

end.
